/*
 *	cook - file construction tool
 *	Copyright (C) 1998 Peter Miller;
 *	All rights reserved.
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
 *
 * MANIFEST: functions to manipulate relati_dirnas
 */

#include <ac/string.h>

#include <builtin/relati_dirna.h>
#include <error_intl.h>
#include <expr/position.h>
#include <str.h>
#include <str_list.h>
#include <trace.h>


/*
 * NAME
 *	relative_dirname - take path apart
 *
 * SYNOPSIS
 *	string_ty *relative_dirname(string_ty *path);
 *
 * DESCRIPTION
 *	relative_dirname is used to extract the directory part
 *	of a pathname.
 *
 * RETURNS
 *	pointer to dynamically allocated string.
 *	A null pointer is returned on error.
 *
 * CAVEAT
 *	Use str_free() when you are done with the value returned.
 */

static string_ty *relative_dirname _((string_ty *));

static string_ty *
relative_dirname(path)
	string_ty	*path;
{
	char		*cp;

	trace(("relative_dirname(path = %08lX)\n{\n"/*}*/, path));
	trace_string(path->str_text);
	cp = strrchr(path->str_text, '/');
	if (cp)
	{
		if (cp > path->str_text)
			path = str_n_from_c(path->str_text, cp - path->str_text);
		else
			path = str_from_c("/");
	}
	else
		path = str_from_c(".");
	trace_string(path->str_text);
	trace(("return %08lX;\n", path));
	trace((/*{*/"}\n"));
	return path;
}


/*
 * NAME
 *	builtin_reldir - relative directory part
 *
 * SYNOPSIS
 *	int builtin_reldir(string_list_ty *result, string_list_ty *args);
 *
 * DESCRIPTION
 *	"reldir" is a built-in function of cook, described as follows:
 *	This function requires one or more arguments, the name of a
 *	files of which to get the dir parts; files relative to the
 *	current directory wil return ".".
 *
 * RETURNS
 *	It returns a string containing the directory parts
 *	of the named files.
 *
 * CAVEAT
 *	The returned result is in dynamic memory.
 *	It is the responsibility of the caller to dispose of
 *	the result when it is finished, with a string_list_destructor() call.
 */

int
builtin_reldir(result, args, pp, ocp)
	string_list_ty	*result;
	const string_list_ty *args;
	const expr_position_ty *pp;
	const struct opcode_context_ty *ocp;
{
	size_t		j;

	trace(("reldir\n"));
	assert(result);
	assert(args);
	assert(args->nstrings);
	if (args->nstrings < 2)
	{
		sub_context_ty	*scp;

		scp = sub_context_new();
		sub_var_set(scp, "Name", "%S", args->string[0]);
		error_with_position
		(
			pp,
			scp,
			i18n("$name: requires one or more arguments")
		);
		sub_context_delete(scp);
		return -1;
	}
	for (j = 1; j < args->nstrings; j++)
	{
		string_ty	*s;

		s = relative_dirname(args->string[j]);
		if (!s)
			return -1;
		string_list_append(result, s);
		str_free(s);
	}
	return 0;
}
